/*
    libgpmi - General Package/Module Interface - lua module
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lua.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "gpmi/debug.h"

/**
 * Returns the extension of this script-module.
 */
char *LUA_GetExtension()
{
	return strdup(".lua");
}

/**
 * Reloads the script on the fly.
 */
int LUA_Rehash(void *client_data)
{
	/* TODO */
	return 0;
}

/**
 * Loads a script.
 */
int LUA_LoadScript(void *client_data, const char *filename)
{
	LUA_ClientData *LUA_MyData = client_data;
	int res;

	/* Load the file */
	res = luaL_loadfile(LUA_MyData->state, filename);

	if (res != 0) {
		/* Error, report it */
		gpmi_error(0, "[LUA] Error: %s", lua_tostring(LUA_MyData->state, -1));
		return -1;
	}

	/* We are supposed to run the file once */
	if (lua_pcall(LUA_MyData->state, 0, 0, 0) != 0) {
	/* Error, report it */
		gpmi_error(0, "[LUA] Error: %s", lua_tostring(LUA_MyData->state, -1));
		return -2;
	}

	return 0;
}

/**
 * Execute a dynamic command.
 **/
static int LUA_DynamicPackageCommand(lua_State *L)
{
	LUA_ClientData *LUA_MyData = GPMI_Script_GetClientData();
	char *(*proc)(int argc, char *argv[]);
	lua_Debug ar;
	char *s;
	int argc, i;
	char **args;

	/* Get the current function-name */
	lua_getstack(L, 0, &ar);
	lua_getinfo(L, "n", &ar);

	proc = GPMI_Script_FindProc(ar.name);

	/* Check if command not found - this must be an internal error: the
	   package had registered the command in the hash, but the routine
	   pointer is NULL.
	*/
	if (proc == NULL) {
		gpmi_debug(0, "[LUA] Internal error: %s has a NULL proc in the command hash\n", ar.name);
		return 0;
	}

	/* Get the amount of args */
	argc = lua_gettop(L);

	/* Allocate memory for the args-array */
	args = malloc((argc + 1) * sizeof(char *));

	args[0] = strdup(ar.name);

	/* Convert everything to strings */
	for (i = 1; i <= argc; i++) {
		args[i] = strdup(lua_tostring(L, i));
	}

	/* Run command */
	s = proc(argc + 1, args);

	/* Free the array */
	for (i = 0; i <= argc; i++)
		free(args[i]);
	free(args);

	if (s == NULL)
		return 0;

	lua_pushstring(LUA_MyData->state, s);

	/* Free result */
	free(s);

	return 1;
}

/**
 * Register a command in our command hash
 */
void LUA_RegisterCommand(void *client_data, const char *name)
{
	LUA_ClientData *LUA_MyData = client_data;

	lua_register(LUA_MyData->state, name, LUA_DynamicPackageCommand);
}

/**
 * Call a function of the script by name, passing params.
 */
void LUA_RunFunction(void *client_data, const char *function, int argc, const char **argv)
{
	LUA_ClientData *LUA_MyData = client_data;
	int i;

	lua_getglobal(LUA_MyData->state, function);

	for (i = 0; i < argc; i++)
		lua_pushstring(LUA_MyData->state, argv[i]);

	lua_call(LUA_MyData->state, argc, 1);
	lua_pop(LUA_MyData->state, 1);
}

int LUA_Initialize(void **client_data)
{
	LUA_ClientData *MyData;

	MyData = malloc(sizeof(LUA_ClientData));
	memset(MyData, 0, sizeof(LUA_ClientData));

	*client_data = MyData;

	/* Initialize a lua interpreter */
	MyData->state = lua_open();

	if (!MyData->state) {
		free(MyData);
		return -1;
	}

	/* LUA doesn't load default libraries by default, so let's do it here */
#if LUA_VERSION_NUM < 500 
	luaopen_base(MyData->state);
	luaopen_io(MyData->state);
	luaopen_string(MyData->state);
	luaopen_math(MyData->state);
#else
	luaL_openlibs(MyData->state);
#endif

	return 0;
}

void LUA_Destroy(void **client_data)
{
	LUA_ClientData *LUA_MyData = *client_data;

	/* Not initialized */
	if (LUA_MyData->state == NULL)
		return;

	lua_close(LUA_MyData->state);

	/* Finally free up the client-data */
	free(*client_data);
}
